class device(object):
    manufacturer = ''
    product_name = ''
    classification = ''
    
    com = None
    _shortcut_command = {}
    
    def __init__(self, com):
        self.com = com
        self.com.open()
        self._add_shortcut_command()
        pass

    def _add_shortcut_command(self):
        items = self._shortcut_command.items()
        for shortcut, method in items:
            self.__setattr__(shortcut,
                             self.__getattribute__(method))
            continue
        return

    def help(self):
        method_list = []
        doc_list = []
        for method in dir(self):
            if method[0]=='_': continue
            if method in self._shortcut_command.keys(): continue
            mt = self.__getattribute__(method)
            if type(mt) in [type(None), int, float, str]: continue
            if mt.__doc__ is None: continue            
            method_help = self.__getattribute__(method).__doc__
            short_help = method_help.split('\n')[1].strip()
            method_list.append(method)
            doc_list.append(short_help)
            continue
        
        length = max(len(m) for m in method_list)
        hlength = max(len(d) for d in doc_list)
        help_fmt = '%-' + '%d'%(length+1) + 's : %s\n'
        separation = '-'*length + '- : ' + '-'*hlength + '\n'
        
        help_str = ''
        help_str += separation
        help_str += help_fmt%('method name', 'description')
        for i in range(len(method_list)):
            if i%20==0: help_str += separation
            help_str += help_fmt%(method_list[i], doc_list[i])
            continue
        help_str += separation
        print(help_str)
        return



from .. import device
import operator

import time

class sp100(device.device):
    manufacturer = 'COSMOTECHS'
    product_name = 'SP100'
    classification = 'Motor Controller'

    def command_set(self, unit, msg):
        target = bytes([32 + unit])
        self._send(target + msg)
        
        d1 = self._recv(1)
        if d1 != b'\x06':
            raise Exception(f'Bad response: {d1}')

        time.sleep(0.1)
        return
    
    def command_read(self, unit, msg):
        target = bytes([32 + unit])
        self._send(target + msg)
        
        d1 = self._recv(1)
        if d1 != b'\x06':
            raise Exception(f'Bad response: {d1}')
        
        d2 = self._readline()
        d3 = self._recv(2)
        bcc = self._calc_bcc(d2[1:-1])
        if d3.upper() != bcc:
            self._send(b'\x15')
            raise Exception(f'BCC not match: receieved = {d3}, calculated = {bcc}')
        
        self._send(b'\x06')
        return d2[4:-1].decode('ascii')
    
    def _send(self, msg):
        if type(msg) == str:
            msg = msg.encode()
            pass
        
        bcc = self._calc_bcc(msg)
        print(b'\x02'+msg+b'\x03'+bcc)
        return self.com.send(b'\x02'+msg+b'\x03'+bcc)

    def _recv(self, num):
        return self.com.recv(num).encode()
    
    def _readline(self):
        terminator = b'\x03'
        
        d = b''
        while True:
            _d = self._recv(1)
            d += _d
            if _d == terminator or _d == b'':
                break
            continue
        return d
        
    def _calc_bcc(self, d):
        tmp = 0
        for _ in d:
            tmp = operator.xor(tmp, _)
            continue
        return f'{tmp:02X}'.encode()
    
    
    # 30h
    # ---

    def origin_return_command(self, ax1=False, ax2=False, ax3=False, ax4=False, unit=0):
        command = b'\x30\x20'
        data = f'{ax1:d},{ax2:d},{ax3:d},{ax4:d}'
        
        return self.command_set(unit, command+data.encode())
    
    def absolute_move_command(self, ax1='', ax2='', ax3='', ax4='', unit=0):
        command = b'\x30\x22'
        data = ''

        if ax1 != '':
            data += f'{ax1:.3f},'
        else:
            data += ','
            pass
        
        if ax2 != '':
            data += f'{ax2:.3f},'
        else:
            data += ','
            pass
        
        if ax3 != '':
            data += f'{ax3:.3f},'
        else:
            data += ','
            pass
        
        if ax4 != '':
            data += f'{ax4:.3f}'
            pass
        
        return self.command_set(unit, command+data.encode())
    
    def immediately_stop_command(self, ax1=False, ax2=False, ax3=False, ax4=False, unit=0):
        command = b'\x30\x24'
        data = f'{ax1:d},{ax2:d},{ax3:d},{ax4:d}'
        
        return self.command_set(unit, command+data.encode())
    
    def decelerate_stop_command(self, ax1=False, ax2=False, ax3=False, ax4=False, unit=0):
        command = b'\x30\x25'
        data = f'{ax1:d},{ax2:d},{ax3:d},{ax4:d}'
        
        return self.command_set(unit, command+data.encode())

    def set_velocity_command(self, ax1=None, ax2=None, ax3=None, ax4=None, unit=0):
        command = b'\x30\x26'
        data = ''

        if ax1 is not None:
            data += f'{ax1:.3f},'
        else:
            data += ','
            pass
        
        if ax2 is not None:
            data += f'{ax2:.3f},'
        else:
            data += ','
            pass
        
        if ax3 is not None:
            data += f'{ax3:.3f},'
        else:
            data += ','
            pass
        
        if ax4 is not None:
            data += f'{ax4:.3f}'
            pass
        
        return self.command_set(unit, command+data.encode())

    def digital_output_query(self, unit=0):
        command = b'\x30\x28'
        
        d = self.command_read(unit, command)
        return [int(_) for _ in f"{int(d, base=16):032b}"[::-1]]
        
    def digital_output_command(self, port, status, unit=0):
        command = b'\x30\x28'
        data = f'{port:d},{status:d}'.encode()
        
        return self.command_set(unit, command+data)
        
    def control_status_query(self, unit=0):
        command = b'\x30\x2a'
        
        d = self.command_read(unit, command)
        bit = [int(_) for _ in f"{int(d, base=16):08b}"[::-1]]
        code = [
            'READY',
            'RUN',
            'ERROR',
            'SPO1',
            'BUSY-1',
            'BUSY-2',
            'BUSY-3',
            'BUSY-4',
        ]
        return {
            'code': d,
            'description': [code[i] for i,b in enumerate(bit) if b == 1],
        }
    
    # 40h
    # ---
    
    def current_positions_query(self, unit=0):
        command = b'\x40\x20'
        
        d = self.command_read(unit, command)
        return [float(_) for _ in d.split(',')]
        
    def axis_status_query(self, unit=0):
        command = b'\x40\x21'
        
        d = self.command_read(unit, command)
        dd = d.split(',')
        descriptions = {
            2**0: 'driving',
            2**1: 'positioning completed',
            2**2: '(NA)',
            2**3: '(NA)',
            2**4: 'origin return not completed',
            2**5: '(NA)',
            2**6: '(NA)',
            2**7: 'anomalous occurrence',
            2**8: 'software limit (+)',
            2**9: 'software limit (-)',
            2**10: 'limit senser (+)',
            2**11: 'limit senser (-)',
            2**12: '(NA)',
            2**13: '(NA)',
            2**14: '(NA)',
            2**15: 'alarm',
        }
        ret = []
        for i, code in enumerate(dd):
            ddd = {
                'code': code,
                'description': [descriptions[2**bit] for bit in range(16)
                                if int(code, base=16) & 2**bit],
            }

            if ddd['description'] == []:
                ddd['description'].append('no error')
                pass
            ret.append(ddd)
            continue

        return ret
            
    def axis_sensers_status_query(self, unit=0):
        command = b'\x40\x23'
        
        d = self.command_read(unit, command)
        dd = d.split(',')
        descriptions = {
            2**0: '+limit',
            2**1: '-limit',
            2**2: 'org',
            2**3: 'z',
            2**4: 'alarm',
            2**5: 'in position',
        }
        ret = []
        for i, code in enumerate(dd):
            ddd = {
                'code': code,
                'description': [descriptions[2**bit] for bit in range(6)
                                if int(code, base=16) & 2**bit],
            }

            if ddd['description'] == []:
                ddd['description'].append('no error')
                pass
            ret.append(ddd)
            continue

        return ret
            
    def error_code_query(self, unit=0):
        command = b'\x40\x24'
        
        d = self.command_read(unit, command)
        descriptions = {
            '00': 'no error',
            '01': '(NA)',
            '02': 'driving command error',
            '03': 'immediately stopped by emergency',
            '04': 'driver error',
            '05': 'program execution error',
            '06': 'origin return not completed',
            '07': 'limit error',
            '08': 'mode error',
            '09': 'positioning error',
            '10': 'command error',
            '11': 'data out of range',
            '12': 'format error',
            '13': 'command out of limit',
            '14': '(NA)',
            '15': '(NA)',
            '16': '(NA)',
            '17': 'normal error',
            '18': '(NA)',
            '19': '(NA)',
            '20': 'flush memory error',
        }
        return {
            'error_code': d,
            'description': descriptions[d],
        }

    def version_query(self, unit=0):
        command = b'\x40\x25'
        
        d = self.command_read(unit, command)
        dd = d.split(',')
        return {
            'version': dd[0],
            'timestamp': f'{dd[1]}-{dd[2]}-{dd[3]}',
        }

    def number_of_axes_query(self, unit=0):
        command = b'\x40\x26'
        
        d = self.command_read(unit, command)        
        descriptions = {
            '0': '2-axes',
            '1': '4-axes',
        }
        return {
            'code': d,
            'description': descriptions[d],
        }
    

    # 50h
    # ---
    
    def change_mode_query(self, unit=0):
        command = b'\x50\x2c'
        
        d = self.command_read(unit, command)
        descriptions = {
            '0': 'parameter',
            '1': 'manual',
            '2': 'position',
            '3': 'program',
            '4': 'auto',
            '5': 'remote',
            '6': 'monitor',
            '7': 'load',
            '8': 'save',
        }
        return {
            'mode': d,
            'description': descriptions[d],
        }
    
    def change_mode_command(self, mode, unit=0):
        command = b'\x50\x2c'
        mode = mode if type(bytes) == str else mode.encode()
        
        self.command_set(unit, command+mode)
        return self.change_mode_query(unit)

    



import time
from . import communicator
from . import ethernet

class gpib_prologix(communicator.communicator):
    method ='gpib_prologix'

    open_flag = False

    host = ''
    gpibport = 10
    lag = 0.02

    def __init__(self, host, gpibport=10, lag=0.02, timeout=10):
        self.gpibport = gpibport
        self.lag = lag
        if type(host) == str:
            self.com = ethernet(host, 1234, timeout)
        else:
            self.com = host
        pass

    def _sleep(self):
        time.sleep(self.lag)
        return

    def open(self):
        if self.open_flag == False:
            self.com.open()
            self.mode_controller()
            self.open_flag = True
            pass
        return

    def close(self):
        self.com.close()
        self.open_flag = False
        return

    def send(self, msg):
        self.use_gpibport()
        self.com.send((msg+self.terminator))
        self._sleep()
        return

    def _send(self, msg):
        self.com.send(msg + self.terminator)
        self._sleep()
        return

    def recv(self, byte):
        self._send('++read %d'%byte)
        ret = self.com.recv(byte)
        return ret

    def readline(self):
        self._send('++read eoi')
        ret = self.com.readline()
        return ret

    def get_info(self):
        self._send('++ver')
        ret = self.readline().strip()
        return ret

    def set_gpibport(self, gpib):
        self.gpibport = int(gpib)
        self.use_gpibport()
        return

    def use_gpibport(self):
        self._send('++addr %d'%(self.gpibport))
        return

    def get_gpibport(self):
        self._send('++addr')
        ret = int(self.readline().strip())
        return ret

    def mode_device(self):
        self._send('++mode 0')
        self.get_mode()
        return

    def mode_controller(self):
        self._send('++mode 1')
        self.get_mode()
        return

    def get_mode(self):
        self._send('++mode')
        ret = int(self.readline().strip())
        return ret



class communicator(object):
    method = 'communicator_base_class'
    connection = False
    
    terminator = '\n'
    
    def __init__(self, *args):
        if len(args)!=0:
            self.open(*args)
            pass
        pass
    
    def set_terminator(self, term_char):
        self.terminator = term_char
        return
        
    def open(self, *args):
        pass
    
    def close(self):
        pass
    
    def send(self, msg):
        pass
    
    def recv(self, byte):
        pass
        
    def readline(self):
        pass





from .. import device

class scpi_common(device.device):
    
    def scpi_accept_address(self):
        """
        *AAD : Accpet Address Command
        -----------------------------
        This command, in conjunction with the Address Set protocol, allows the
        controller to detect all address-configurable devices (that is, devices
        that implement this command) and assign an IEEE 488.1 address to each of
        those devices. An address-configurable device is detected when the
        controller has completed a byte-by-byte search of the device's
        identifier, after which time the address is assigned. The Address Set
        protocol causes this identifier search to be executed repeatedly until
        all address-configurable devices have been detected.
        """
        self.com.send('*AAD')
        return
        
    def scpi_calibration_query(self):
        """
        *CAL : Calibration Query
        ------------------------
        The Calibration query causes a device to perform an internal
        self-calibration and generate a response that indicates whether or not
        the device completed the self-calibration without error. Additional
        """
        self.com.send('*CAL?')
        ret = self.com.readline()
        ret = int(ret)
        return ret
     
    def scpi_clear_status(self):
        """
        *CLS : Clear Status Command
        ---------------------------
        The Clear Status command clears status data structures, and forces the 
        device to the Operation Complete Command Idle State and the Operation 
        Complete Query Idle State.
        """
        self.com.send('*CLS')
        return
    
    def scpi_define_device_trigger(self):
        """
        *DDT : Define Device Trigger Command
        ------------------------------------
        The Define Device Trigger command stores a command sequence that is
        executed when a group execute trigger (GET), IEEE 488.1 interface
        message, or *TRG common command is received. The *RST common command
        shall set the command sequence to a device-defined state.
        """
        self.com.send('*DDT')
        return
        
    def scpi_define_device_trigger_query(self):
        """
        *DDT? : Define Device Trigger Query
        -----------------------------------
        The Define Device Trigger query allows the programmer to examine the
        command sequence which will be executed when a GET or *TRG command is
        received.
        """
        self.com.send('*DDT?')
        ret = self.com.readline()
        ret = ret.strip()
        return ret

    def scpi_disable_listener_function(self):
        """
        *DLF : Disable Listener Function
        --------------------------------
        The Disable Listener Function command causes a device to cease being a
        listener (change to L0 subset). If this command is the first
        device-specific message received after the device leaves IEEE 488.1 DCAS
        state, the device shall cease being a listener within 100 ms after the
        acceptance of the <PROGRAM MESSAGE TERMINATOR>. A subsequent DCL message
        shall restore listener capability. The device shall resume listening
        within 100 ms after entering IEEE 488.1 DCAS state.
        """
        self.com.send('*DLF')
        return
        
    def scpi_define_macro(self):
        """
        *DMC : Define Macro Command
        ---------------------------
        The Define Macro command allows the programmer to assign a sequence of
        zero or more <PROGRAM MESSAGE UNIT> elements to a macro label.
        The sequence is executed when the label is received as a
        <COMMAND PROGRAM HEADER> or <QUERY PROGRAM HEADER>.
        """
        self.com.send('*DMC')
        return
        
    def scpi_enable_macro(self):
        """
        *EMC : Enable Macro Command
        ---------------------------
        The Enable Macro command enables and disables expansion of macros.
        Macro definitions are not affected by this command. One use of this
        command is to turn off macro expansion in order to execute a
        device-specific command with the same name as a macro. The *RST command
        disables the expansion of macros.
        """
        self.com.send('*EMC')
        return
        
    def scpi_enable_macro_query(self):
        """
        *EMC? : Enable Macro Query
        --------------------------
        The Enable Macro query allows the programmer to query whether the macros
        are enabled. A returned value of zero indicates that macros are
        disabled. A returned value of one indicates that macros are enabled.
        """
        self.com.send('*EMC?')
        ret = self.com.readline()
        ret = int(ret)
        return ret
        
    def scpi_standard_event_status_enable(self):
        """
        *ESE : Standard Event Status Enable Command
        -------------------------------------------
        The Standard Event Status Enable command sets the Standard Event Status
        Enable Register bits.
        """
        self.com.send('*ESE')
        return
        
    def scpi_standard_event_status_enable_query(self):
        """
        *ESE? : Standard Event Status Enable Query
        ------------------------------------------
        The Standard Event Status Enable query allows the programmer to
        determine the current contents of the Standard Event Status Enable
        Register.
        """
        self.com.send('*ESE?')
        ret = self.com.readline()
        ret = int(ret)
        return ret
        
    def scpi_standard_event_status_register_query(self):
        """
        *ESR : Standard Event Status Register Query
        -------------------------------------------
        The Standard Event Status Register query allows the programmer to
        determine the current contents of the Standard Event Status Register.
        Reading the Standard Event Status Register clears it.
        """
        self.com.send('*ESR?')
        ret = self.com.readline()
        ret = int(ret)
        return ret
        
    def scpi_get_macro_contents_query(self):
        """
        *GMC? : Get Macro Contents Query
        --------------------------------
        The Get Macro Contents query allows the current definition of a macro to
        be retrieved from a device.
        """
        self.com.send('*GMC?')
        ret = self.com.readline()
        ret = ret.strip()
        return ret
        
    def scpi_identification_query(self):
        """
        *IDN? : Identification Query
        ----------------------------
        The intent of the Identification query is for the unique identification
        of devices over the system interface.
        """
        self.com.send('*IDN?')
        ret = self.com.readline()
        ret = ret.strip().split(',')
        return ret
    
    def scpi_individual_status_query(self):
        """
        *IST? : Individual Status Query
        -------------------------------
        The Individual Status query allows the programmer to read the current
        state of the IEEE 488.1 defined "ist" local message in the device.
        """
        self.com.send('*IST?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def scpi_learn_macro_query(self):
        """
        *LMC? : Learn Macro Query
        -------------------------
        This query returns the currently defined macro labels.
        """
        self.com.send('*LMC?')
        ret = self.com.readline()
        ret = ret.strip()
        return ret

    def scpi_learn_device_setup_query(self):
        """
        *LRN? : Learn Device Setup Query
        --------------------------------
        The Learn Device Setup query allows the programmer to obtain a sequence
        of <RESPONSE MESSAGE UNIT> elements that may later be used as
        <PROGRAM MESSAGE UNIT> elements to place the device in the state it was
        in when the *LRN? common query was made.
        """
        self.com.send('*LRN?')
        self.com.readline()
        ret = ret.strip()
        return ret
        
    def scpi_operation_complete(self):
        """
        *OPC : Operation Complete Command
        ---------------------------------
        The Operation Complete command causes the device to generate the
        operation complete message in the Standard Event Status Register
        when all pending selected device operations have been finished.
        """
        self.com.send('*OPC')
        return

    def scpi_operation_complete_query(self):
        """
        *OPC? : Operation Complete Query
        --------------------------------
        The Operation Complete query places an ASCII character "1" into the
        device's Output Queue when all pending selected device operations have
        been finished.
        """
        self.com.send('*OPC?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def scpi_option_identification_query(self):
        """
        *OPT? : Option Identification Query
        -----------------------------------
        The Option Identification query is for identifying reportable device
        options over the system interface.
        """
        self.com.send('*OPT?')
        ret = self.com.readline()
        ret = ret.strip()
        return ret

    def scpi_pass_control_back(self):
        """
        *PCB : Pass Control Back
        ------------------------
        The Pass Control Back command is used by a controller to tell a device,
        being a potential controller, to which address the control is to be
        passed back when the device (acting as a controller) sends the
        IEEE 488.1 interface message, take control (TCT).
        """
        self.com.send('*PCB')
        return

    def scpi_purge_macros(self):
        """
        *PMC : Purge Macros Command
        ---------------------------
        The Purge Macros command causes the device to delete all macros that
        may have been previously defined using the *DMC command. All stored
        macro command sequences and labels shall be removed from the device's
        memory by this command.
        """
        self.com.send('*PMC')
        return

    def scpi_parallel_poll_enable_register(self):
        """
        *PRE : Parallel Poll Enable Register Command
        --------------------------------------------
        The Parallel Poll Enable Register command sets the Parallel Poll
        Enable Register bits.
        """
        self.com.send('*PRE')
        return

    def scpi_parallel_poll_enable_register_query(self):
        """
        *PRE? : Parallel Poll Enable Register Query
        -------------------------------------------
        The Parallel Poll Enable Register query allows the programmer to
        determine the current contents of the Parallel Poll Enable Register.
        """
        self.com.send('*PRE?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def scpi_power_on_status_clear_query(self):
        """
        *PSC? : Power-On Status Clear Query
        -----------------------------------
        The Power-On Status Clear query allows the programmer to query
        the device's power-on-status-clear flag. A returned value of zero
        indicates that the Standard Event Status Enable Register, Service
        Request Enable Register, and the Parallel Poll Enable Register will
        retain their status when power is restored to the device. A returned
        value of one indicates that the registers listed above will be cleared
        when power is restored to the device.
        """
        self.com.send('*PSC?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def scpi_protected_user_data(self):
        """
        *PUD : Protected User Data Command
        ----------------------------------
        The Protected User Data command stores data unique to the device such
        as calibration date, usage time, environmental conditions, and inventory
        control numbers. A minimum of 63 bytes shall be provided. The size of
        this area shall be specified in the device documentation.
        """
        self.com.send('*PUD')
        return

    def scpi_protected_user_data_query(self):
        """
        *PUD? : Protected User Data Query
        ---------------------------------
        The Protected User Data query allows the programmer to retrieve
        the contents of the *PUD storage area. See *PUD command.
        """
        self.com.send('*PUD?')
        ret = self.com.readline()
        ret = ret.strip()
        return ret        
        return

    def scpi_recall(self):
        """
        *RCL : Recall Command
        ---------------------
        The *RCL command restores the current settings of a device from a
        copy stored in local memory. The scope of the *RCL command is the
        same as *RST and the *LRN? response. Device documentation shall
        explicitely mention the device settings that are restored by *RCL.
        """
        self.com.send('*RCL')
        return

    def scpi_resource_description_transfer(self):
        """
        *RDT : Resource Description Transfer Command
        --------------------------------------------
        The Resource Description Transfer command allows a Resource
        Description to be stored in a device.
        """
        self.com.send('*RDT')
        return

    def scpi_resource_description_transfer_query(self):
        """
        *RDT? : Resource Description Transfer Query
        -------------------------------------------
        The Resource Description Transfer query allows a Resource Description
        to be retrieved from a device. The Resource Description may be memory
        or in a read-write memory settable by the *RDT command.
        """
        self.com.send('*RDT?')
        ret = self.com.readline()
        ret = ret.strip()
        return ret

    def scpi_reset(self):
        """
        *RST : Reset Command
        --------------------
        The Reset command performs a device reset. The Reset command is the
        third level of reset in a three-level reset strategy.

        The Reset command shall do the following:

          1) Except as explicitly excluded below, set the device-specific
             functions to a known state that is independent of the
             past-use history of the device. Device-specific commands may be
             provided to program a different reset state than the original
             factory-supplied one.
          2) Set the macro defined by *DDT to a device-defined state.
          3) Disable macros.
          4) Force the device into the OCIS state.
          5) Force the device into the OQIS state.
        
        The reset command explicitly shall NOT affect the following:
        
          1) The state of the IEEE 488.1 interface.
          2) The selected IEEE 488.1 address of the device.
          3) The Output Queue.
          4) Any Event Enable Register setting, including the Standard Event
             Status Enable Register settings.
          5) Any Event Register setting, including the Standard Event Status
             Register settings,
          6) The power-on-status-clear flag setting.
          7) Macros defined with the Define Macro Contents command.
          8) Calibration data that affects device specifications.
          9) The Protected User Data query response.
          10) The Resource Description Transfer query response.
          11) The Service Request Enable Register setting.
          12) The Parallel Poll Enable Register setting.
          13) The memory register(s) associated with *SAV.
        """
        self.com.send('*RST')
        return

    def scpi_save(self):
        """
        *SAV : Save Command
        -------------------
        The *SAV command stores the current settings of the device in local
        memory. The scope of the *SAV command is the same as *RST and the
        *LRN? response. Device documentation shall explicitly mention the
        device settings that are restored by *RCL.
        """
        self.com.send('*SAV')
        return

    def scpi_service_request_enable(self, bit):
        """
        *SRE : Service Request Enable Command
        -------------------------------------
        The Service Request Enable command sets the Service Request Enable
        Register bits.
        """
        self.com.send('*SRE %d'%(bit))
        return

    def scpi_service_request_enable_query(self, bit):
        """
        *SRE? : Service Request Enable Query
        ------------------------------------
        The Service Request Enable query allows the programmer to determine
        the current contents of the Service Request Enable Register.
        """
        self.com.send('*SRE?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def scpi_read_status_byte_query(self):
        """
        *STB? : Read Status Byte Query
        -----------------------------
        The Read Status Byte query allows the programmer to read the status
        byte and Master Summary Status bit.
        """
        self.com.send('*STB?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def scpi_trigger(self):
        """
        *TRG : Trigger Command
        ----------------------
        The Trigger command is the device-specific analog of the IEEE 488.1
        defined Group Execute Trigger (GET) interface message, and has exactly
        the same effect as a GET when received, parsed, and executed by the
        device.
        """
        self.com.send('*TRG')
        return

    def scpi_self_test_query(self):
        """
        *TST? : Self-Test Query
        ----------------------
        The self-test query causes an internal self-test and places a response
        into the Output Queue indicating whether or not the device completed
        the self-test without any detected errors. Optionally, information
        on why the self-test was not completed may be contained in the
        response.
        """
        self.com.send('*TST?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def scpi_wait_to_continue(self):
        """
        *WAI : Wait-to-Continue Command
        -------------------------------
        The Wait-to-Continue command shall prevent the device from executing
        any further commands or queries until the no- operation-pending flag
        is TRUE.
        """
        self.com.send('*WAI')
        return

    def scpi_remove_individual_macro(self):
        """
        *RMC : Remove Individual Macro Command
        --------------------------------------
        The Remove Individual Macro command removes a single macro definition
        from the device.
        """
        self.com.send('*RMC')
        return

    def scpi_save_default_device_settings(self):
        """
        *SDS : Save Default Device Settings Command
        -------------------------------------------
        The Save Default Device Settings command initializes the contents
        of a save/recall register. The register contents are set to a known
        state that is independent of the past use history of the device.
        """
        self.com.send('*SDS')
        return
    
    _scpi_dict = {'*AAD': 'scpi_accept_address',
                  '*CAL?': 'scpi_calibration_query',
                  '*CLS': 'scpi_clear_status',
                  '*DDT': 'scpi_define_device_trigger',
                  '*DDT?': 'scpi_define_device_trigger_query',
                  '*DLF': 'scpi_disable_listener_function',
                  '*DMC': 'scpi_define_macro',
                  '*EMC': 'scpi_enable_macro',
                  '*EMC?': 'scpi_enable_macro_query',
                  '*ESE': 'scpi_standard_event_status_enable',
                  '*ESE?': 'scpi_standard_event_status_enable_query',
                  '*ESR?': 'scpi_standard_event_status_register_query',
                  '*GMC?': 'scpi_get_macro_contents_query',
                  '*IDN?': 'scpi_identification_query',
                  '*IST?': 'scpi_individual_status_query',
                  '*LMC?': 'scpi_learn_macro_query',
                  '*LRN?': 'scpi_learn_device_setup_query',
                  '*OPC': 'scpi_operation_complete',
                  '*OPC?': 'scpi_operation_complete_query',
                  '*OPT?': 'scpi_option_identification_query',
                  '*PCB': 'scpi_pass_control_back',
                  '*PMC': 'scpi_purge_macros',
                  '*PRE': 'scpi_parallel_poll_enable_register',
                  '*PRE?': 'scpi_parallel_poll_enable_register_query',
                  '*PSC?': 'scpi_power_on_status_clear_query',
                  '*PUD': 'scpi_protected_user_data',
                  '*PUD?': 'scpi_protected_user_data_query',
                  '*RCL': 'scpi_recall',
                  '*RDT': 'scpi_resource_description_transfer',
                  '*RDT': 'scpi_resource_description_transfer_query',
                  '*RST': 'scpi_reset',
                  '*SAV': 'scpi_save',
                  '*SRE': 'scpi_service_request_enable',
                  '*SRE?': 'scpi_service_request_enable_query',
                  '*STB?': 'scpi_read_status_byte_query',
                  '*TRG?': 'scpi_trigger',
                  '*TST?': 'scpi_self_test_query',
                  '*WAI': 'scpi_wait_to_continue',
                  '*RMC': 'scpi_remove_individual_macro',
                  '*SDS': 'scpi_save_default_device_settings'}


class scpi_family(device.device):

    _scpi_enable = 'ALL'

    def __init__(self, com):
        device.device.__init__(self, com)
        self._scpi = scpi_common(com)
        self._add_scpi_methods()
        
    def _add_scpi_methods(self):
        sdic = self._scpi._scpi_dict
        if self._scpi_enable=='ALL': add = sdic.items()
        else:
            add = []
            for enable in self._scpi_enable.split(' '):
                add.append([enable, sdic[enable]])
                continue
            pass
        for call, method in add:
            self.__setattr__(method,
                             self._scpi.__getattribute__(method))
            shortcut = call.replace('*', '').replace('?', 'Q')
            self.__setattr__(shortcut, self.__getattribute__(method))
            continue
        return
    



#!/usr/bin/env python3

import time
import serial
import struct

class tr_73u(object):

    def __init__(self, port):
        self.ser = self.serial_open(port)
        self.start()
        return

    def serial_open(self, port):
        ser = serial.Serial(port=port, baudrate=19200, timeout=5, stopbits=1)
        return ser

    def start(self):
        self.write(b'0x00')
        return

    def write(self, cmd):
        self.ser.write(cmd)
        return

    def output_current_data(self):
        data = self.query(cmd=b'\x01\x33\x00\x04\x00\x00\x00\x00\x00\x38\x00', byte=26)
        #data = 26byte, format=B
        data = struct.unpack('26B', data)
        ret = {}
        ret['temp'] = (data[6]*16**2+data[5]-1000)/10
        ret['humid'] = (data[8]*16**2+data[7]-1000)/10
        ret['press'] = (data[10]*16**2+data[9])/10
        return ret

    def query(self, cmd, byte, wt=0.01):
        self.write(cmd)
        time.sleep(wt)
        ret = self.read(byte)
        return ret

    def read(self, byte=100):
        ret = self.ser.read(byte)
        return ret



import sys
import time
import socket
import numpy
import ogameasure
from ..SCPI import scpi


# ==============
# Helper Classes
# ==============

# Bias Hex Changer
# ================
class bias_changer(object):
    min = 0
    max = 1

    def __init__(self, bias=None, hex=None):
        self.bins = numpy.linspace(self.min, self.max, 0x1000)
        self.th = (self.bins + (abs(self.bins[1]-self.bins[0])/2.))[:-1]
        if bias is not None:
            self.bias = bias
            self.digit = numpy.digitize([bias], self.th)[0]
        elif hex is not None:
            if type(hex) == str: self.digit = int(hex, 16)
            elif type(hex) == int: self.digit = hex
            self.bias = self.bins[self.digit]
        else:
            raise ValueError
        self.hexstr = '%04X'%(self.digit)
        pass

class bias_changer_100(bias_changer):
    min = 0
    max = 100

class bias_changer_200(bias_changer):
    min = 0
    max = 200


# Error Class
# ===========
class error_item(object):
    num = 0
    msg = ''
    txt = ''

    def __init__(self, num, msg, txt):
        self.num = num
        self.msg = msg
        self.txt = txt
        pass

class error_handler(object):
    error_list = [
        error_item(-100, 'Command error', ''),
        error_item(-200, 'Execution error', ''),
        error_item(-400, 'Query error', ''),
        error_item(-350, 'Queue overflow', ''),
    ]

    @classmethod
    def check(cls, num, msg):
        if num==0: return
        for e in cls.error_list:
            if msg == e.msg:
                emsg = '%s (%d)'%(e.msg, e.num)
                msg = 'Power meter returned Error message.\n'
                msg += '*'*len(emsg) + '\n'
                msg += emsg + '\n'
                msg += '*'*len(emsg) + '\n'
                msg += e.txt + '\n'
                raise StandardError(msg)
            continue
        _msg = 'Power meter returned Error message.\n'
        emsg = '%s (%d)\n'%(msg, num)
        _msg += '*'*len(emsg) + '\n'
        _msg += emsg
        _msg += '*'*len(emsg) + '\n'
        raise StandardError(_msg)
        return


# ==========
# main class
# ==========

class GPDVC15(scpi.scpi_family):
    manufacturer = 'ELVA-1'
    product_name = 'GPDVC-15'
    classification = 'Attenuator Driver'

    _scpi_enable = '*IDN? *SAV'

    _bias_changer = bias_changer

    def gpib_address_search(self):
        print('GPIB Address Searching...')
        for i in range(31):
            sys.stdout.write('\r try address %d ...   '%(i))
            sys.stdout.flush()
            com = ogameasure.gpib_prologix(self.com.com.host, i)
            try:
                com.com.timeout = 1.2
                com.open()
                com.send('*IDN?')
                com.readline()
            except socket.timeout:
                com.close()
                sys.stdout.write('\x08\x08NG')
                sys.stdout.flush()
                time.sleep(0.1)
                continue
            sys.stdout.write('\x08\x08OK\n')
            sys.stdout.flush()
            com.close()
            return i
        return

    def _error_check(self):
        err_num, err_msg = self.error_query()
        error_handler.check(err_num, err_msg)
        return

    def error_query(self):
        """
        SYST:ERR? : Query Error Numbers
        -------------------------------
        Querry error number.

        Args
        ====
        Nothing.

        Returns
        =======
        < err_num : int :  >
            Error number. 0 = 'No Error'

        < err_msg : str :  >
            Error message.

        Examples
        ========
        >>> a.error_query()
        (0, 'No error.')
        """
        self.com.send('SYST:ERR?')
        ret = self.com.readline()
        ret = ret.strip().split(',')
        err_num = int(ret[0])
        err_msg = ret[1].strip('"')
        return err_num, err_msg

    def version_query(self):
        """
        SYST:VERS : Query Version
        -------------------------
        Querry Version

        Args
        ====
        Nothing.

        Returns
        =======
        < version : float :  >
            Version of the system.

        Examples
        ========
        >>> a.version_query()
        1994.0
        """
        self.com.send('SYST:VERS?')
        ret = self.com.readline()
        self._error_check()
        ret = float(ret)
        return ret

    def gpib_address_set(self, gpib_address):
        """
        SYST:COMM:GPIB:ADDR : Set GPIB Address
        --------------------------------------
        Set GPIB Address.

        NOTE : After executed this command, reconnect the GPIB connection
               to continue communications.

        NOTE : To save the changed configurations, please execute '*SAV 0'
               command.

        Args
        ====
        < gpib_address : int :  >
            Specify the GPIB address to set.

        Returns
        =======
        Nothing.

        Examples
        ========
        >>> a.gpib_address_set(4)
        """
        self.com.send('SYST:COMM:GPIB:ADDR %d'%(gpib_address))
        return

    def gpib_address_query(self):
        """
        SYST:COMM:GPIB:ADDR? : Query GPIB Address
        -----------------------------------------
        Query GPIB Address.

        Args
        ====
        Nothing.

        Returns
        =======
        < gpib_address : int :  >
            GPIB address.

        Examples
        ========
        >>> a.gpib_address_set(4)
        """
        self.com.send('SYST:COMM:GPIB:ADDR?')
        ret = self.com.readline()
        self._error_check()
        ret = int(ret)
        return ret

    def output_set(self, output):
        """
        PO : Set Output Current
        -----------------------
        Set the output current by mA.

        Args
        ====
        < output : float : (mA) >
            Specify the bias current to output.
            unit is mA

        Returns
        =======
        Nothing.

        Examples
        ========
        >>> a.output_set(0)
        >>> a.output_set(10.1)
        """
        output = self._bias_changer(bias=output)
        self.com.send('PO %s'%(output.hexstr))
        self.output = output
        return

    def output_get(self):
        """
        Get the output current.

        Args
        ====
        Nothing.

        Returns
        =======
        < output : float : (mA) >
            Bias current. Unit is mA.

        Examples
        ========
        >>> a.output_get()
        10.234
        """
        return self.output.bias


class GPDVC15_100(GPDVC15):
    _bias_changer = bias_changer_100

class GPDVC15_200(GPDVC15):
    _bias_changer = bias_changer_200



#! /usr/bin/env python3

import time, numpy

class tpg261_lan(object):

    def __init__(self, com):
        self.com = com
        self.com.open()
        pass

    def pressure(self):
        self.com.send("PR1 \r\n")
        time.sleep(0.1)
        self.com.send("\x05")
        time.sleep(0.1)
        self.raw_p = self.com.recv().decode('utf-8').strip().split(',')
        time.sleep(0.1)
        pressure = float(self.raw_p[-1])
        return pressure

    def tpg_status(self):
        self.com.send(b"PR1 \r\n")
        time.sleep(0.3)
        self.com.send(b"\x05")
        time.sleep(0.3)
        self.get = self.com.recv()
        status_p = self.get[3:4]
        if status_p == b'0' :
            print('Measurement data okay')
        elif status_p == b'1' :
            print('Underrange')
        elif status_p == b'2' :
            print('Overrange')
        elif status_p == b'3' :
            print('Sensor error')
        elif status_p == b'4' :
            print('Sensor off (IKR, PKR, IMR, PBR)')
        elif status_p == b'5' :
            print('No sensor')
        else :
             print('Identification error')

    def turn_status_g1(self):
        self.com.send(b"SEN , 0, 0 \r\n")
        time.sleep(0.3)
        self.com.send(b"\x05")
        time.sleep(0.3)
        self.get = self.com.recv()
        status1_g = self.get[3:4]
        if status1_g == b'0' :
            print('Gauge cannot be turned on/off')
        elif status1_g == b'1' :
            print('Gauge turned off')
        elif status1_g == b'2' :
            print('Gauge turned on')
        else :
            pass

    def gauge_on_g1(self):
        self.com.send(b"SEN , 2 , 0 \r\n")
        time.sleep(0.3)
        self.com.send(b"\x05")
        time.sleep(0.3)
        self.get = self.com.recv()
        status1_g = self.get[3:4]
        if status1_g == b'0' :
            print('Gauge cannot be turned on/off')
        elif status1_g == b'1' :
            print('Gauge turned off')
        elif status1_g == b'2' :
            print('Gauge turned on')
        else :
            pass

    def gauge_off_g1(self):
        self.com.send(b"SEN , 1 , 0 \r\n")
        time.sleep(0.3)
        self.com.send(b"\x05")
        time.sleep(0.3)
        self.get = self.com.recv()
        status1_g = self.get[3:4]
        if status1_g == b'0' :
            print('Gauge cannot be turned on/off')
        elif status1_g == b'1' :
            print('Gauge turned off')
        elif status1_g == b'2' :
            print('Gauge turned on')
        else :
            pass

    def pres_unit_bar(self):
        self.com.send(b"UNI,0 \r\n")
        time.sleep(0.3)
        self.com.send(b"\x05")
        time.sleep(0.3)
        self.raw_bar = self.com.recv()
        unit = self.raw_bar[3:4]
        return unit

    def pres_unit_torr(self):
        self.com.send(b"UNI,1 \r\n")
        time.sleep(0.3)
        self.com.send(b"\x05")
        time.sleep(0.3)
        self.raw_torr = self.com.recv()
        unit = self.raw_torr[3:4]
        return unit

    def pres_unit_pa(self):
        self.com.send(b"UNI,2 \r\n")
        time.sleep(0.3)
        self.com.send(b"\x05")
        time.sleep(0.3)
        self.raw_pa = self.com.recv()
        unit = self.raw_pa[3:4]
        return unit



import sys
import datetime
from ..SCPI import scpi
import serial

class model218_usb(object):
    manufacturer = 'Lakeshore'
    product_name = 'model 218'
    classification = 'Temperature Monitor'

    _scpi_enable = '*CLS *ESE *ESE? *ESR? *IDN? *OPC *OPC? *RST *SRE ' \
                   + '*SRE? *STB? *TST? *WAI'

    def __init__(self, port):
        port = port
        self.com = serial.Serial(port=port,
                                baudrate=9600,
                                bytesize=7,
                                parity="O",
                                stopbits=1,
                                timeout=3,
                                write_timeout=3)
        pass

    def send(self,cmd):
        terminator = "\r\n"
        _cmd = cmd + terminator
        cmd_byte = _cmd.encode()
        self.com.write(cmd_byte)

    def alarm_set(self, ch=1, off_on=1, source=1, high_value=300,
                  low_value=100, deadband=1, latch_enable=0):
        """
        ALARM : Configure Input Alarm Parameters
        ----------------------------------------
        Configures the alarm parameters for an input.

        < ch : int : 1-8 >
            Specifies which input to configure (1-8).

        < off_on : int : 0,1 >
            Determines whether the instrument checks the alarm for this input.
            0 = OFF, 1 = ON

        < source : int : 1-4 >
            Specifies input data to check.
            1 = Kelvin, 2 = Celsius, 3 = sensor units, 4 = linear data.

        < high_value : float :  >
            Sets the value the source is checked against to activate the
            high alarm.

        < low_value : float :  >
            Sets the value the source is checked against to activate low
            alarm.

        < deadband : float :  >
            Sets the value that the source must change outside of an alarm
            condition to deactivate an unlatched alarm.

        < latch_enable : int : >
            Specifies a latched alarm (remains active after alarm condition
            correction).
        """
        self.send('ALARM %d, %d, %d %.3f, %.3f, %.3f, %d'%
                      (ch, off_on, source, high_value, low_value, deadband,
                       latch_enable))
        return

    def alarm_query(self, ch=1, printlog=True):
        """
        ALARM? : Query Input Alarm Parameters
        -------------------------------------
        Returns the alarm parameters of an input.

        < ch : int : 1-8 >
            Specifies which input to configure (1-8).

        Returns:
            off_on, source, high_value, low_value, deadband, latch_enable
        """
        self.send('ALARM? %d'%(ch))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        off_on = int(ret[0])
        source = int(ret[1])
        high_value = float(ret[2])
        low_value = float(ret[3])
        deadband = float(ret[4])
        latch_enable = int(ret[5])
        if printlog:
            print('ALARM? %d'%(ch))
            print('---------')
            print('off/on = %d'%(off_on))
            print('source = %d'%(source))
            print('high_value = %.3f'%(high_value))
            print('low_value = %.3f'%(low_value))
            print('deadband = %.3f'%(deadband))
            print('latch_enable = %d'%(latch_enable))
            print('')
            pass
        return off_on, source, high_value, low_value, deadband, latch_enable

    def alarm_query_all(self, printlog=True):
        """
        (Helper Method) ALARM? : Query Input Alarm Parameters
        -----------------------------------------------------
        Returns the alarm parameters of an input.

        Returns:
            off_on, source, high_value, low_value, deadband, latch_enable
        """
        off_on = []
        source = []
        high_value = []
        low_value = []
        deadband = []
        latch_enable = []
        for i in range(1, 9):
            ret = self.alarm_query(i, printlog=False)
            off_on.append(ret[0])
            source.append(ret[1])
            high_value.append(ret[2])
            low_value.append(ret[3])
            deadband.append(ret[4])
            latch_enable.append(ret[5])
            continue
        if printlog:
            sep = ''
            print('ALARM?')
            print('------')
            print('#   : ON/OFF, Sourse,    High,     Low, DeadBand, Latch')
            for i in range(8):
                print('ch%d :      %d,      %d, %07.3f, %07.3f,  %07.3f, %d'%
                      (i+1, off_on[i], source[i], high_value[i], low_value[i],
                       deadband[i], latch_enable[i]))
                continue
            print('')
            pass
        return off_on, source, high_value, low_value, deadband, latch_enable

    def alarm_status_query(self, ch=1, printlog=True):
        """
        ALARMST? : Query Input Alarm Status
        -----------------------------------
        Returns the alarm status of an input.

        < ch : int : 1-8 >
            Specifies which input to configure (1-8).

        Returns:
            high_status : Specifies high alarm status.
                          0 = Unactivated, 1 = Activated

            low_status : Specifies low alarm status.
                         0 = Unactivated, 1 = Activated
        """
        self.send('ALARMST? %d'%(ch))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        high_status = int(ret[0])
        low_status = int(ret[1])
        if printlog:
            print('ALARMST? %d'%(ch))
            print('-----------')
            print('high_status = %d'%(high_status))
            print('low_status = %d'%(low_status))
            print('')
            pass
        return high_status, low_status

    def alarm_status_query_all(self, printlog=True):
        """
        (Helper Method) ALARMST? : Query Input Alarm Status
        ---------------------------------------------------
        Returns the alarm status of an input.

        Returns:
            high_status : Specifies high alarm status.
                          0 = Unactivated, 1 = Activated

            low_status : Specifies low alarm status.
                         0 = Unactivated, 1 = Activated
        """
        high_status = []
        low_status = []
        for i in range(1, 9):
            ret = self.alarm_status_query(i, printlog=False)
            high_status.append(ret[0])
            low_status.append(ret[1])
            continue
        if printlog:
            print('ALARMST?')
            print('--------')
            print('#   : High, Low')
            for i in range(8):
                print('ch%d :    %d,  %d'%(i+1, high_status[i], low_status[i]))
                continue
            print('')
            pass
        return high_status, low_status

    def audible_alarm_set(self, off_on=0):
        """
        ALMB : Configure Audible Alarm
        ------------------------------
        Enables or disables system alarm beeper.

        < off_on : int : 0,1 >
            disables/enables beeper.
            1 = On, 0 = Off
        """
        self.send('ALMB %d'%(off_on))
        return

    def audible_alarm_query(self):
        """
        ALMB? : Query Audible Alarm
        -----------------------------------
        Returns system beeper parameters.

        Returns:
            off_on : disables/enables beeper.
                     1 = On, 0 = Off
        """
        self.send('ALMB?')
        ret = self.com.readline()
        ret = int(ret)
        print('ALMB?')
        print('-----')
        print('off_on = %d'%(ret))
        print('')
        return ret

    def alarm_reset(self):
        """
        ALMRST : Clear Alarm Status for All Inputs
        ------------------------------------------
        Resets a latched active alarm after the alarm condition has cleared.
        """
        self.send('ALMRST')
        return

    def analog_outputs_set(self, ch=1, bipolar_enable=0, mode=1, monitor_ch=1,
                           source=1, high_value=200, low_value=2, manual=0):
        """
        ANALOG : Configure Analog Output Parameters
        -------------------------------------------
        Configure Analog Output Parameters.

        < ch : int : 1,2 >
            Specifies which analog output to configure (1 or 2).

        < bipolar_enable : int : 0,1 >
            Specifies analog output:
            0 = positive only, 1 = bipolar.

        < mode : int : 0,1,2 >
            Specifies data the analog output monitors:
            0 = off, 1 = input, 2 = manual.

        < monitor_ch : int : 1-8 >
            Specifies which input to monitor if <mode> = 1 (1-8).

        < source : int : 1-4 >
            Specifies input data.
            1 = Kelvin, 2 = Celsius, 3 = sensor units, 4 = linear equation.

        < high_value : float :  >
            If <mode> = 1, this parameter represents the data at which the
            analog output reaches +100% output.

        < low_value : float :  >
            If <mode> = 1, this parameter represents the data at which the
            analog output reaches -100% output if bipolar, or 0% output if
            positive only.

        < manual : float :  >
            If <mode> = 2, this parameter is the output of the analog output.
        """
        self.send('ANALOG %d, %d, %d, %d, %d, %.3f, %.3f, %.3f'%
                      (ch, bipolar_enable, mode, monitor_ch, source,
                       high_value, low_value, manual))
        return

    def analog_outputs_query(self, ch=1, printlog=True):
        """
        ANALOG? : Query Analog Output Parameters
        ----------------------------------------

        < ch : int : 1,2 >
            Specifies analog output to query (1 or 2).

        Returns:
            bipolar_enable, mode, monitor_ch, source, high_value,
            low_value, manual_value
        """
        self.send('ANALOG? %d'%(ch))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        bipolar_enable = int(ret[0])
        mode = int(ret[1])
        monitor_ch = int(ret[2])
        source = int(ret[3])
        high_value = float(ret[4])
        low_value = float(ret[5])
        manual = float(ret[6])
        if printlog:
            print('ANALOG? %d'%(ch))
            print('----------')
            print('bipolar_enable = %d'%(bipolar_enable))
            print('mode = %d'%(mode))
            print('monitor_ch = %d'%(monitor_ch))
            print('source = %d'%(source))
            print('high_value = %.3f'%(high_value))
            print('low_value = %.3f'%(low_value))
            print('manual = %.3f'%(manual))
            print('')
            pass
        return (bipolar_enable, mode, monitor_ch, source, high_value,
                low_value, manual)

    def analog_output_data_query(self, ch=1):
        """
        AOUT? : Query Analog Output Data
        --------------------------------
        Returns the percentage of output.

        < ch : int : 1,2 >
            Specifies analog output to query (1 or 2).

        Returns:
            output
        """
        self.send('AOUT? %d'%(ch))
        ret = self.com.readline()
        ret = float(ret)
        return ret

    def serial_interface_baud_rate_set(self, bps=2):
        """
        BAUD : Configure Serial Interface Baud Rate.
        --------------------------------------------
        Configures to serial interface baud rate.

        < bps : int : 0,1,2 >
            Specifies bits per second (bps) rate.
            0 = 300, 1 = 1200, 2 = 9600.
        """
        self.send('BAUD %d'%(bps))
        return

    def serial_interface_baud_rate_query(self):
        """
        BAUD? : Query Serial Interface Baud Rate.
        -----------------------------------------
        Returns serial interface baud rate.

        Returns:
            bps : 0 = 300, 1 = 1200, 2 = 9600.
        """
        self.send('BAUD?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def celsius_reading_query(self, ch=0):
        """
        CRDG? : Query Celsius Reading for Single Input or All Inputs
        ------------------------------------------------------------
        Returns the Celsius reading for a single input or all inputs.

        < ch : int : 0-8 >
            0 = all inputs. Return a list of float values.
            1-8 = individual input. Return a float value.

        NOTE: Use 0 (all inputs) when reading two or more inputs
              at the maximum update rate of 16 rdg/s.
        """
        self.send('CRDG? %d'%(ch))
        ret = self.com.readline()
        ret = map(float, ret.strip().split(','))
        if ch!=0: ret = ret[0]
        return ret

    def curve_delete(self, curve_num):
        """
        CRVDEL : Delete User Curve
        --------------------------
        Deletes a user curve.

        < curve_num : int : 21-28 >
            Specifies which curve to delete (21-28) for inputs 1-8.
        """
        self.send('CRVDEL %d'%(curve_num))
        return

    def curve_header_set(self, curve_num, name, SN, format, limit_value,
                         coefficient):
        """
        CRVHDR : Configure Curve Header
        -------------------------------

        < curve_num : int : 21-28 >
            Specifies which curve to configure (21-28) for inputs 1-8.

        < name : str[15] : >
            Specifies curve name. Limited to 15 characters.

        < SN : str[10] :  >
            Specifies curve serial number. Limited to 10 characters.

        < format : int : 2-4 >
            Specifies curve data format.
            2 = V/K, 3 = Ohm/K, 4 = log Ohm/K

        < limit_value : float : >
            Specifies curve temperature limit in Kelvin.

        < coefficient : int : 1,2 >
            Specifies curve temperature coefficient.
            1 = negative, 2 = positive.
        """
        self.send('CRVHDR %d, %s, %s, %d, %.3f, %d'%
                      (curve_num, name, SN, format, limit_value, coefficient))
        return

    def curve_header_query(self, curve_num, printlog=True):
        """
        CRVHDR? : Configure Curve Header
        --------------------------------

        < curve_num : int : 21-28 >
            Specifies which curve to configure (21-28) for inputs 1-8.

        Returns:
            name, SN, format, limit_value, coefficient
        """
        self.send('CRVHDR? %d'%(curve_num))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        name = ret[0].strip()
        SN = ret[1].strip()
        format = int(ret[2])
        limit_value = float(ret[3])
        coefficient = int(ret[4])
        if printlog:
            print('CRVHDR? %d'%(curve_num))
            print('----------')
            print('name = %s'%(name))
            print('SN = %s'%(SN))
            print('format = %d'%(format))
            print('limit_value = %.3f'%(limit_value))
            print('coefficient = %d'%(coefficient))
            print('')
            pass
        return name, SN, format, limit_value, coefficient

    def curve_header_query_all(self, printlog=True):
        """
        (Helper Method) CRVHDR? : Configure Curve Header
        ------------------------------------------------

        Returns:
            name, SN, format, limit_value, coefficient
        """
        name = []
        SN = []
        format = []
        limit_value = []
        coefficient = []
        for i in range(1, 10) + range(21, 29):
            ret = self.curve_header_query(i, printlog=False)
            name.append(ret[0])
            SN.append(ret[1])
            format.append(ret[2])
            limit_value.append(ret[3])
            coefficient.append(ret[4])
            continue
        if printlog:
            print('CRVHDR?')
            print('-------')
            print('#       : %15s, %10s, %6s, %7s, %5s'%
                  ('name', 'serial #', 'format', 'limit', 'coeff'))
            for i, num in enumerate(range(1, 10) + range(21, 29)):
                print('curve%02d : %15s, %10s, %6d, %7.3f,  %5d'%
                      (num, name[i], SN[i], format[i], limit_value[i],
                       coefficient[i]))
                continue
            print('')
            pass
        return name, SN, format, limit_value, coefficient

    def curve_point_set(self, curve_num, index, units_value, temp_value):
        """
        CRVPT : Configure Curve Data Point
        ----------------------------------
        Configures a user curve data point.

        < curve_num : int : 21-28 >
            Specifies which curve to configure (21-28) for inputs 1-8.

        < index : int : 1-200 >
            Specifies the points index in the curve (1 - 200).

        < units_value : float : >
            Specifies sensor units for this point.

        < temp_value : float : >
            Specifies corresponding temperature in Kelvin for this point.
            6 digits.
        """
        units = ('%f'%(units_value))[:7]
        temp = ('%f'%(temp_value))[:7]
        self.send('CRVPT %d %d %s %s'%(curve_num, index, units, temp))
        return

    def curve_point_query(self, curve_num, index):
        """
        CRVPT? : Query Curve Data Point
        -------------------------------
        Returns a standard or user curve data point.

        < curve_num : int : 1-9,21-28 >
            Specifies which curve to query.
            1-5 = Standard Diode Curves,
            6-9 = Standard Platinum Curves,
            21-28 = User Curves.
            NOTE: Curve locations 10-20 not used.

        < index : int : 1-200 >
            Specifies the points index in the curve (1 - 200).

        Returns:
            units_value, temp_value
        """
        self.send('CRVPT? %d %d'%(curve_num, index))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        units_value = float(ret[0])
        temp_value = float(ret[1])
        return units_value, temp_value

    def curve_point_set_line(self, curve_num, units_values, temp_values):
        """
        (Helper Method) CRVPT : Configure Curve Data Point
        --------------------------------------------------
        Configures a user curve data point at one time.

        < curve_num : int : 21-28 >
            Specifies which curve to configure (21-28) for inputs 1-8.

        < units_values : list : >
            A list of sensor units for the curve.
            Max length is 200 (first 200 items will be used in excess case).

        < temp_value : list : >
            A list of corresponding temperatures in Kelvin for the curve.
            Max length is 200 (first 200 items will be used in excess case).
        """
        print('uploading...')
        for i,(unit, temp) in enumerate(zip(units_values, temp_values)):
            if i>199: break
            sys.stdout.write('\r[%-10s] %d/200 %.1f %.1f'%
                             ('='*int(i/19), i+1, unit, temp))
            sys.stdout.flush()
            self.curve_point_set(curve_num, i+1, unit, temp)
            continue
        print('')
        return

    def curve_point_query_line(self, curve_num):
        """
        (Helper Method) CRVPT? : Query Curve Data Point
        -----------------------------------------------
        Returns a standard or user curve data as a 1-d data.

        < curve_num : int : 1-9,21-28 >
            Specifies which curve to query.
            1-5 = Standard Diode Curves,
            6-9 = Standard Platinum Curves,
            21-28 = User Curves.
            NOTE: Curve locations 10-20 not used.

        Returns:
            units_values, temp_values
        """
        unit = []
        temp = []
        print('downloading...')
        for i in range(200):
            _u, _t = self.curve_point_query(curve_num, i+1)
            unit.append(_u)
            temp.append(_t)
            sys.stdout.write('\r[%-10s] %d/200 %.1f %.1f'%
                             ('='*int(i/19), i+1, _u, _t))
            sys.stdout.flush()
            continue
        print('')
        return unit, temp

    def datetime_set(self, year, month, day, hour, minutes, seconds):
        """
        DATETIME : Configure Date and Time
        ----------------------------------
        Configures date and time using 24-hour format.

        < year : int : 00-99 >
            Specifies year. Valid entries are: 00 - 99.

        < month : int : 1-12 >
            Specifies month. Valid entries are: 1 - 12.

        < day : int : 1-31 >
            Specifies day. Valid entries are 1 - 31.

        < hour : int : 0-23 >
            Specifies hour. Valid entries are: 0 - 23.

        < minutes : int : 0-59 >
            Specifies minutes. Valid entries are: 0 - 59.

        < seconds : int : 0-59 >
            Specifies seconds. Valid entries are: 0 - 59.
        """
        self.send('DATETIME %d, %d, %d, %d, %d, %d'%
                      (month, day, year, hour, minutes, seconds))
        return

    def datetime_set_now(self):
        """
        (Helper Method) DATETIME : Configure Date and Time
        --------------------------------------------------
        Configures date and time to current time.
        """
        fmt = '%m,%d,%y,%H,%M,%S'
        datetime_str = datetime.datetime.now().strftime(fmt)
        self.send('DATETIME %s'%(datetime_str))
        return

    def datetime_query(self):
        """
        DATETIME? : Query Date and Time
        --------------------------------
        Returns date and time.

        Returns:
            timestamp
        """
        self.send('DATETIME?')
        ret = self.com.readline().strip()
        fmt = '%m,%d,%y,%H,%M,%S'
        timestamp = datetime.datetime.strptime(ret, fmt)
        return timestamp

    def factory_defaults_reset(self, param=0):
        """
        DFLT : Set to Factory Defaults
        ------------------------------
        Sets all configuration values to factory defaults and resets the
        instrument. Does not clear user curves or instrument calibration.

        < param : int : 99 >
            The 99 is required to prevent accidentally setting the
            unit to defaults.
        """
        self.send('DFLT %d'%(param))
        return

    def display_field_set(self, location, sensor, source):
        """
        DISPFLD : Configure Display Parameters
        --------------------------------------
        Configures the display parameters.

        < location : int : 1-8 >
            Specifies display location to configure (1-8).

        < sensor : int : 0-8 >
            Specifies input to display in the display location (0-8).(0=none).

        < source : int : 1-4 >
            Specifies input data to display.
            1 = Kelvin, 2 = Celsius, 3 = sensorunits, 4 = linear data,
            5 = minimum data, 6 = maximum data.
        """
        self.send('DISPFLD %d, %d, %d'%(location, sensor, source))
        return

    def display_field_set_all_kelvin(self):
        """
        (Helper Method) DISPFLD : Configure Display Parameters
        ------------------------------------------------------
        Set all display fields to show Kelvin value.
        """
        for i in range(1,9):
            self.display_field_set(i, i, 1)
            continue
        return

    def display_field_query(self, location):
        """
        DISPFLD? : Query Display Field
        ------------------------------
        Returns the parameters for a displayed field.

        < location : int : 1-8 >
            Specifies display location to configure (1-8).

        Returns:
            sensor, source
        """
        self.send('DISPFLD? %d'%(location))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        sensor = int(ret[0])
        source = int(ret[1])
        return sensor, source

    def display_field_query_all(self, printlog=True):
        """
        (Helper Method) DISPFLD? : Query Display Field
        -----------------------------------------------
        Returns the parameters for all displayed fields.

        Returns:
            sensors, sources
        """
        sensors = []
        sources = []
        for i in range(1,9):
            _se, _so = self.display_field_query(i)
            sensors.append(_se)
            sources.append(_so)
            continue
        if printlog:
            print('DISPFLD?')
            print('--------')
            print('#         : %6s, %6s'%('sensor', 'source'))
            for i in range(8):
                print('display%02d : %6d, %6d'%(i+1, sensors[i], sources[i]))
                continue
            print('')

        return sensors, sources

    def filter_set(self, ch, off_on=1, points=5, window=2):
        """
        FILTER : Configure Input Filter Parameters
        ------------------------------------------

        < ch : int : 1-8 >
            Specifies input to configure (1-8).

        < off_on : int : 0,1 >
            Specifies whether the filter function is off or on.
            0 = Off, 1 = On.

        < points : int : 2-64 >
            Specifies how many data points the filtering function uses (2-64).

        < window : int : 1-10 >
            Specifies what percent of full scale reading limits the
            filtering function (1-10). Reading changes greater than this
            percentage reset the filter.
        """
        self.send('FILTER %d, %d, %d, %d'%(ch, off_on, points, window))
        return

    def filter_set_all(self, off_on=1, points=5, window=2):
        """
        (Helper Method) FILTER : Configure Input Filter Parameters
        ----------------------------------------------------------
        Set same configurations for all input sensors.

        < off_on : int : 0,1 >
            Specifies whether the filter function is off or on.
            0 = Off, 1 = On.

        < points : int : 2-64 >
            Specifies how many data points the filtering function uses (2-64).

        < window : int : 1-10 >
            Specifies what percent of full scale reading limits the
            filtering function (1-10). Reading changes greater than this
            percentage reset the filter.
        """
        for i in range(1, 9):
            self.filter_set(i, off_on, points, window)
            continue
        return

    def filter_query(self, ch):
        """
        FILTER? : Query Input Filter Parameters
        ---------------------------------------

        < ch : int : 1-8 >
            Specifies input to configure (1-8).

        Returns:
            off_on, points, window
        """
        self.send('FILTER? %d'%(ch))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        off_on = int(ret[0])
        points = int(ret[1])
        window = int(ret[2])
        return off_on, points, window

    def filter_query_all(self, printlog=True):
        """
        (Helper Method) FILTER? : Query Input Filter Parameters
        -------------------------------------------------------
        Query input filter parameters for all input sensors.

        Returns:
            off_on, points, windows
        """
        off_on = []
        points = []
        window = []
        for i in range(1, 9):
            _o, _p, _w = self.filter_query(i)
            off_on.append(_o)
            points.append(_p)
            window.append(_w)
            continue
        if printlog:
            print('FILTER?')
            print('-------')
            print('#   : %6s, %6s, %6s'%('ON_OFF', 'points', 'window'))
            for i in range(8):
                print('ch%d : %6d, %6d, %6d'%
                      (i+1, off_on[i], points[i], window[i]))
                continue
            print('')
            pass
        return off_on, points, window

    def ieee488_set(self, terminator=0, eoi=0, address=30):
        """
        IEEE : Configure IEEE-488 Interface Parameters
        ----------------------------------------------
        Configures parameters of the IEEE interface.

        < terminator : int : 0-3 >
            Specifies the terminator.
            0 = <CR><LF>, 1 = <LF><CR>, 2 = <LF>, 3 = no terminator.

        < eoi : int : 0,1 >
            Disables/enables the EOI mode.
            0 = Enabled, 1 = Disabled.

        < address : int : >
            Specifies the IEEE address.
        """
        self.send('IEEE %d, %d, %d'%(terminator, eoi, address))
        return

    def ieee488_query(self):
        """
        IEEE? : Query IEEE-488 Interface Parameters
        -------------------------------------------
        Returns IEEE interface parameters.

        Returns:
            terminator, eoi, address
        """
        self.send('IEEE?')
        ret = self.com.readline()
        ret = ret.strip().split(',')
        terminator = int(ret[0])
        eoi = int(ret[1])
        address = int(ret[2])
        return terminator, eoi, address

    def input_curve_set(self, ch, curve_number):
        """
        INCRV : Configure Input Curve Number
        ------------------------------------
        Specifies the curve an input uses for temperature conversion.

        < ch : int : 1-8 >
            Specifies which input to configure (1-8).

        < curve_number : int : 0-9,21-28 >
            Specifies which curve the input uses.
            0 = none, 1-5 = Standard Diode Curves,
            6-9 = Standard Platinum Curves, 21-28 = User curves.
            Note: Curve locations 10-20 not used.
        """
        self.send('INCRV %d %d'%(ch, curve_number))
        return

    def input_curve_query(self, ch):
        """
        INCRV? : Query Input Curve Number
        ---------------------------------
        Returns the input curve number.

        < ch : int : 1-8 >
            Specifies which input to configure (1-8).

        Returns:
            curve_number
        """
        self.send('INCRV? %d'%(ch))
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def input_curve_query_all(self, printlog=True):
        """
        (Helper Method) INCRV? : Query Input Curve Number
        -------------------------------------------------
        Returns the list of input curve numbers.

        Returns:
            curve_numbers
        """
        curves = [self.input_curve_query(i) for i in range(1, 9)]
        if printlog:
            print('INCRV?')
            print('------')
            print('#   : %7s'%('curve #'))
            for i in range(8):
                print('ch%d : %7d'%(i+1, curves[i]))
                continue
            print('')
            pass
        return curves

    def input_control_set(self, ch, off_on):
        """
        INPUT : Configure Input Control Parameter
        -----------------------------------------
        Turns selected input on or off.

        < ch : int : 1-8 >
            Specifies which input to configure(1-8).

        < off_on : int : 0,1 >
            Disables/Enables input.
            0 = Off, 1 = On.
        """
        self.send('INPUT %d %d'%(ch, off_on))
        return

    def input_control_query(self, ch):
        """
        INPUT? : Query Input Control Parameter
        --------------------------------------
        Returns selected input status.

        < ch : int : 1-8 >
            Specifies which input to configure(1-8).

        Returns:
            input_stauts
        """
        self.send('INPUT? %d'%(ch))
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def input_control_query_all(self, printlog=True):
        """
        (Helper Method) INPUT? : Query Input Control Parameter
        ------------------------------------------------------
        Returns all input status.

        Returns:
            input_stauts
        """
        status = [self.input_control_query(i) for i in range(1, 9)]
        if printlog:
            print('INPUT?')
            print('------')
            print('#   : %6s'%('status'))
            for i in range(8):
                print('ch%d : %6d'%(i+1, status[i]))
                continue
            print('')
            pass
        return status

    def input_type_set(self, group, sensor_type):
        """
        INTYPE : Configure Input Type Parameters
        ----------------------------------------
        Configures input type parameters for a group of inputs.

        < group : str[1] : A,B >
            Specifies input group to configure.
            A = inputs 1-4, B = inputs 5-8.

        < sensor_type : int :  >
            Specifies input sensor type. Valid entries:
            0 = 2.5V Diode, 1 = 7.5V Diode, 2 = 250ohm Platinum,
            3 = 500ohm Platinum, 4 = 5kohm Platinum, 5 = Cernox
        """
        self.send('INTYPE %s, %d'%(group, sensor_type))
        return

    def input_type_query(self, group):
        """
        INTYPE? : Query Input Type Parameters
        -------------------------------------
        Returns input type parameters.

        < group : str[1] : A,B >
            Specifies input group to configure.
            A = inputs 1-4, B = inputs 5-8.

        Returns:
            sensor_type
        """
        self.send('INTYPE? %s'%(group))
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def input_type_query_all(self):
        """
        (Helper Method) INTYPE? : Query Input Type Parameters
        -----------------------------------------------------
        Returns list of input type parameters.

        Returns:
            sensor_types
        """
        sensor_types = [self.input_type_query(g) for g in ('A', 'B')]
        return sensor_types

    def keypad_status_query(self):
        """
        KEYST? : Query Keypad Status
        ----------------------------
        Returns keypad status since the last KEYST?.

        Returns:
            status : 1 = key pressed,
                     0 = no key pressed.
                     NOTE: KEYST? returns 1 after initial power-up.
        """
        self.send('KEYST?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def kelvin_reading_query(self, ch=0):
        """
        KRDG? : Query Kelvin Reading for Single Input or All Inputs
        -----------------------------------------------------------
        Returns the Kelvin reading for a single input or all inputs.

        < ch : int : 0-8 >
            0 = all inputs. Return a list of float values.
            1-8 = individual input. Return a float value.

        NOTE: Use 0 (all inputs) when reading two or more inputs
              at the maximum update rate of 16 rdg/s.
        """
        self.send('KRDG? %d '%(ch))
        ret = self.com.readline()
        ret = map(float, ret.decode('utf-8').strip().split(','))
        if ch!=0: ret = ret[0]
        return ret

    def linear_equation_set(self, ch, m, source, b):
        """
        LINEAR : Configure Input Linear Equation Parameters
        ---------------------------------------------------
        Configures the linear equation for an input.
        y = m * x(source) + b

        < ch : int : 1-8 >
            Specifies input to configure (1-8).

        < m : float :  >
            Specifies a value for m in the equation.

        < source : int : 1-3 >
            Specifies input data.
            1 = Kelvin, 2 = Celsius, 3 = sensor units.

        < b : float :  >
            Specifies a value for b in the equation.
        """
        self.send('LINEAR %d, %.3f, %d, %.3f'%(ch, m, source, b))
        return

    def linear_equation_query(self, ch):
        """
        LINEAR? : Query Input Linear Equation Parameters
        ------------------------------------------------
        Returns input linear equation configuration.
        y = m * x(source) + b

        < ch : int : 1-8 >
            Specifies input to configure (1-8).

        Returns:
            m, source, b
        """
        self.send('LINEAR? %d'%(ch))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        m = float(ret[0])
        source = int(ret[1])
        b = float(ret[2])
        return m, source, b

    def linear_equation_query_all(self, printlog=True):
        """
        (Helper Method) LINEAR? : Query Input Linear Equation Parameters
        ---------------------------------------------------------------
        Returns input linear equation configuration for all ch.
        y = m * x(source) + b

        Returns:
            m, source, b
        """
        m = []
        source = []
        b = []
        for i in range(1, 9):
            _m, _s, _b = self.linear_equation_query(i)
            m.append(_m)
            source.append(_s)
            b.append(_b)
            continue
        if printlog:
            print('LINEAR?')
            print('-------')
            print('#   : %7s, %6s, %7s'%('m', 'source', 'b'))
            for i in range(8):
                print('ch%d : %7.3f, %6d, %7.3f'%
                      (i+1, m[i], source[i], b[i]))
                continue
            print('')
            pass
        return m, source, b

    def lockout_set(self, off_on=0, code=123):
        """
        LOCK : Configure Lock-out and Lock-out Code
        -------------------------------------------
        Configures keypad lock-out and lock-out code.

        < off_on : int : 0,1 >
            Disables/enables the keypad lock-out.
            0 = OFF, 1 = ON

        < code : int : 0-999 >
            Specifies lock-out code. 000 - 999.
        """
        self.send('LOCK %d %d'%(off_on, code))
        return

    def lockout_query(self):
        """
        LOCK? : Query Lock-out and Lock-out Code
        ----------------------------------------
        Returns lock-out status and lock-out code.

        Returns:
            off_on, code
        """
        self.send('LOCK?')
        ret = self.com.readline()
        ret = ret.strip().split(',')
        off_on = int(ret[0])
        code = int(ret[1])
        return off_on, code

    def logging_on_off(self, off_on):
        """
        LOG : Turns Logging On and Off
        ------------------------------
        Turns logging on and off.

        < off_on : int : 0,1 >
            0 = Off, 1 = On.
        """
        self.send('LOG %d'%(off_on))
        return

    def logging_on_off_query(self):
        """
        LOG? : Query Logging Status
        ---------------------------
        Returns logging status.

        Returns:
            off_on
        """
        self.send('LOG?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def logging_number_query(self):
        """
        LOGNUM? : Query Number of Last Data Log Record Stored
        ----------------------------------------------------
        Returns number of last data log record stored.

        Returns:
            log_num
        """
        self.send('LOGNUM?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def logging_records_set(self, reading_number, ch, source):
        """
        LOGREAD : Configure Log Records
        -------------------------------
        Configures log records.

        < reading_number : int : 1-8 >
            The individual reading number (1-8) within a log record to
            configure.

        < ch : int : 1-8 >
            The input number to log (1-8).

        < source : int : 1-4 >
            Specifies data source to log.
            1 = Kelvin, 2 = Celsius, 3 = sensor units, 4 = linear data.
        """
        self.send('LOGREAD %d, %d, %d'%(reading_number, ch, source))
        return

    def logging_records_query(self, reading_number):
        """
        LOGREAD? : Query Log Records
        ----------------------------
        Returns log record parameters.

        < reading_number : int : 1-8 >
            The individual reading number (1-8) within a log record to
            configure.

        Returns:
            ch, source
        """
        self.send('LOGREAD? %d'%(reading_number))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        ch = int(ret[0])
        source = int(ret[1])
        return ch, source

    def logging_records_query_all(self, printlog=True):
        """
        (Helper Method) LOGREAD? : Query Log Records
        --------------------------------------------
        Returns log record parameters for all records.

        Returns:
            ch, source
        """
        ch = []
        source = []
        for i in range(1, 9):
            _c, _s = self.logging_records_query(i)
            ch.append(_c)
            source.append(_s)
            continue
        if printlog:
            print('LOGREAD?')
            print('--------')
            print('#   : %6s, %6s'%('sensor', 'source'))
            for i in range(8):
                print('ch%d : %6d, %6d'%
                      (i+1, ch[i], source[i]))
                continue
            print('')
            pass
        return ch, source

    def logging_parameter_set(self, mode, overwrite, start, period, readings):
        """
        LOGSET : Configure Logging Parameters
        -------------------------------------
        Configures logging parameters.

        < mode : int : 0-4 >
            Specifies logging mode.
            0 = Off, 1 = Log Continuous, 2 = Log event,
            3 = Print Continuous, 4 = Print Event.

        < overwrite : int : 0,1 >
            Specifies overwrite mode.
            0 = Do not overwrite data, 1 = overwrite data.

        < start : int : 0,1 >
            Specifies start mode. 0 = Clear, 1 = Continue.

        < period : int : 1-3600 >
            Specifies period in seconds (1-3600).
            If mode is Print Continuous, minimum period is 10.

        < readings : int : 1-8 >
            Specifies number of readings per record (1-8).
        """
        self.send('LOGSET %d, %d, %d, %d, %d'%(mode, overwrite, start,
                                                   period, readings))
        return

    def logging_parameter_query(self):
        """
        LOGSET? : Query Logging Parameters
        ----------------------------------
        Returns logging parameters.

        Returns:
            mode, overwrite, start, period, readings
        """
        self.send('LOGSET?')
        ret = self.com.readline()
        ret = ret.strip().split(',')
        mode = int(ret[0])
        overwrite = int(ret[1])
        start = int(ret[2])
        period = int(ret[3])
        readings = int(ret[4])
        return mode, overwrite, start, period, readings

    def log_data_query(self, record_number, reading_number):
        """
        LOGVIEW? : Querya Logged Data Record
        ------------------------------------
        Returns a single reading from a logged data record.

        < record_number : int :  >

        < reading_number : int : 1-8 >
            The individual reading number (1-8) within a log record to
            configure.

        Returns:
            datetime, reading, status, source
        """
        self.send('LOGVIEW? %d, %d'%(record_number, reading_number))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        date = ret[0]
        time = ret[1]
        fmt = '%m/%d/%y%H:%M:%S'
        datime = datetime.datetime.strptime(date+time, fmt)
        reading = float(ret[2])
        status = int(ret[3])
        source = int(ret[4])
        return datime, reading, status, source

    def linear_equation_input_data_query(self, ch=0):
        """
        LRDG? : Query Linear Equation Data for a Single Input or All Inputs.
        --------------------------------------------------------------------
        Returns the linear equation data for an input.

        < ch : int : 0-8 >
            0 = all inputs. Return a list of float values.
            1-8 = individual input. Return a float value.

        NOTE: Use 0 (all inputs) when reading two or more inputs
              at the maximum update rate of 16 rdg/s.
        """
        self.send('LRDG? %d'%(ch))
        ret = self.com.readline()
        ret = map(float, ret.strip().split(','))
        if ch!=0: ret = ret[0]
        return ret

    def minmax_set(self, ch, source):
        """
        MNMX : Configure Minimum and Maximum Input Function Parameters
        --------------------------------------------------------------
        Configures the minimum and maximum input functions.

        < ch : int : 1-8 >
            Specifies input to configure (1-8).

        < source : int : 1-4 >
            Specifies input data to process through max/min.
            1 = Kelvin, 2 = Celsius, 3 = sensor units, 4 = linear data.
        """
        self.send('MNMX %d %d'%(ch, source))
        return

    def minmax_query(self, ch):
        """
        MNMX? : Query Minimum and Maximum Input Function Parameters
        -----------------------------------------------------------
        Returns an input min/max configuration.

        < ch : int : 1-8 >
            Specifies input to configure (1-8).

        Returns:
            source
        """
        self.send('MNMX? %d'%(ch))
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def minmax_data_query(self, ch):
        """
        MNMXRDG? : Query Min/Max Data for an Input
        ------------------------------------------
        Returns the minimum and maximum input data.

        < ch : int : 1-8 >
            Specifies input to configure (1-8).

        Returns:
            min_value, max_value
        """
        self.send('MNMXRDG? %d'%(ch))
        ret = self.com.readline()
        ret = ret.strip().split(',')
        min_value = float(ret[0])
        max_value = float(ret[1])
        return min_value, max_value

    def minmax_function_reset(self):
        """
        MNMXRST : Resets Min/Max Function for All Inputs
        ------------------------------------------------
        Resets the minimum and maximum data for all inputs.
        """
        self.send('MNMXRST')
        return

    def local_remote_mode_set(self, mode):
        """
        MODE : Configure Remote Interface Mode
        --------------------------------------
        Configures the remote interface mode.

        < mode : int : 0-2 >
            specifies which mode to operate.
            0 = local, 1 = remote, 2 = remote with local lockout.
        """
        self.send('MODE %d'%(mode))
        return

    def local_remote_mode_query(self):
        """
        MODE? : Query Remote Interface Mode
        -----------------------------------
        Returns the remote interface mode.

        Returns:
            mode
        """
        self.send('MODE?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def reading_status_query(self, ch):
        """
        RDGST? : Query Input Status
        ---------------------------
        The integer returned represents the sum of the bit weighting of the
        input status flag bits.

        < ch : int : 1-8 >
            specifies which input to query.

        Returns:
            status
        """
        self.send('RDGST? %d'%(ch))
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def relay_set(self, relay_number, mode, input_alarm, alarm_type):
        """
        RELAY : Configure Relay Control Parameters
        ------------------------------------------
        Configures relay control.

        < relay_number : int : 1-8 >
            Specifies which relay to configure (1-8).

        < mode : int : 0-2 >
            Specifies relay mode. 0 = Off, 1 = On, 2 = Alarms

        < input_alarm : int : 1-8 >
            Specifies which input alarm activates the relay when the relay
            is in alarm mode (1-8).

        < alarm_type : int : 0-2 >
            Specifies the input alarm type that activates the relay when the
            relay is in alarm mode.
            0 = Low alarm, 1 = High Alarm, 2 = Both Alarms.
        """
        self.send('RELAY %d, %d, %d, %d'%(relay_number, mode,
                                              input_alarm, alarm_type))
        return

    def relay_query(self, relay_number):
        """
        RELAY? : Query Relay Control Parameters
        ---------------------------------------
        The integer returned represents the sum of the bit weighting of the
        relay status.

        < relay_number : int : 1-8 >
            Specifies which relay to configure (1-8).

        Returns:
            mode, input_alarm, alarm_type
        """
        self.send('RELAY? %d'%(relay_number))
        ret = self.com.readline()
        ret = self.com.readline()
        ret = ret.strip().split(',')
        mode = int(ret[0])
        input_alarm = int(ret[1])
        alarm_type = int(ret[2])
        return mode, input_alarm, alarm_type

    def relay_status_query(self):
        """
        RELAYST? : Query Relay Status
        -----------------------------
        The integer returned represents the sum of the bit weighting of the
        relay status.

        Returns:
            status
        """
        self.send('RELAYST?')
        ret = self.com.readline()
        ret = int(ret)
        return ret

    def softcal_curve_generate(self, std, dest, SN, T1, U1, T2, U2, T3, U3):
        """
        SCAL : Generate SoftCal Curve
        -----------------------------

        < std : int : 1,6,7 >
        < dest : int : 21-28 >
        < SN : str[10] :  >
        < T1 : float : >
        < U1 : float : >
        < T2 : float : >
        < U2 : float : >
        < T3 : float : >
        < U3 : float : >
        """
        self.send('SCAL %d, %d, %s, %.3f, %.3f, %.3f, %.3f, %.3f, %.3f'%
                      (std, dest, SN, T1, U1, T2, U2, T3, U3))
        return

    def sensor_units_reading_query(self, ch=0):
        """
        SRDG? : Query Sensor Units Reading for Single Input or All Inputs
        -----------------------------------------------------------------
        Returns the Sensor Units reading for a single input or all inputs.

        < ch : int : 0-8 >
            0 = all inputs. Return a list of float values.
            1-8 = individual input. Return a float value.

        NOTE: Use 0 (all inputs) when reading two or more inputs
              at the maximum update rate of 16 rdg/s.
        """
        self.send('SRDG? %d'%(ch))
        ret = self.com.readline()
        ret = map(float, ret.strip().split(','))
        if ch!=0: ret = ret[0]
        return ret

    _shortcut_command = {'ALARM': 'alarm_set',
                         'ALARMQ': 'alarm_query',
                         'ALARMSTQ': 'alarm_status_query',
                         'ALMB': 'audible_alarm_set',
                         'ALMBQ': 'audible_alarm_query',
                         'ALMRST': 'alarm_reset',
                         'ANALOG': 'analog_outputs_set',
                         'ANALOGQ': 'analog_outputs_query',
                         'AOUTQ': 'analog_output_data_query',
                         'BAUD': 'serial_interface_baud_rate_set',
                         'BAUDQ': 'serial_interface_baud_rate_query',
                         'CRDGQ': 'celsius_reading_query',
                         'CRVDEL': 'curve_delete',
                         'CRVHDR': 'curve_header_set',
                         'CRVHDRQ': 'curve_header_query',
                         'CRVPT': 'curve_point_set',
                         'CRVPTQ': 'curve_point_query',
                         'DATETIME': 'datetime_set',
                         'DATETIMEQ': 'datetime_query',
                         'DFLT': 'factory_defaults_reset',
                         'DISPFLD': 'display_field_set',
                         'DISPFLDQ': 'display_field_query',
                         'FILTER': 'filter_set',
                         'FILTERQ': 'filter_query',
                         'IEEE': 'ieee488_set',
                         'IEEEQ': 'ieee488_query',
                         'INCRV': 'input_curve_set',
                         'INCRVQ': 'input_curve_query',
                         'INPUT': 'input_control_set',
                         'INPUTQ': 'input_control_query',
                         'INTYPE': 'input_type_set',
                         'INTYPEQ': 'input_type_query',
                         'KEYSTQ': 'keypad_status_query',
                         'KRDGQ': 'kelvin_reading_query',
                         'LINEAR': 'linear_equation_set',
                         'LINEARQ': 'linear_equation_query',
                         'LOCK': 'lockout_set',
                         'LOCKQ': 'lockout_query',
                         'LOG': 'logging_on_off',
                         'LOGQ': 'logging_on_off_query',
                         'LOGNUMQ': 'logging_number_query',
                         'LOGREAD': 'logging_records_set',
                         'LOGREADQ': 'logging_records_query',
                         'LOGSET': 'logging_parameter_set',
                         'LOGSETQ': 'logging_parameter_query',
                         'LOGVIEWQ': 'log_data_query',
                         'LRDGQ': 'linear_equation_input_data_query',
                         'MNMX': 'minmax_set',
                         'MNMXQ': 'minmax_query',
                         'MNMXRDGQ': 'minmax_data_query',
                         'MNMXRST': 'minmax_function_reset',
                         'MODE': 'local_remote_mode_set',
                         'MODEQ': 'local_remote_mode_query',
                         'RDGSTQ': 'reading_status_query',
                         'RELAY': 'relay_set',
                         'RELAYQ': 'relay_query',
                         'RELAYSTQ': 'relay_status_query',
                         'SCAL': 'softcal_curve_generate',
                         'SRDGQ': 'sensor_units_reading_query'}



from .SCPI import scpi_common

from . import Agilent
from . import Agilent as Keysight
from . import Anritsu
#from . import Canon
from . import Cosmotechs
from . import ELVA1
from . import HEIDENHAIN
from . import KIKUSUI
from . import Lakeshore
from . import Pfeiffer
from . import Phasematrix
from . import SENA
from . import TandD
